/*Standard return errors
#define S_OK			0x00000000L
#define E_OUTOFMEMORY	0x80000002L
#define E_FAIL			0x80004005L*/

//Standard Types
typedef unsigned long DWORD;
typedef unsigned char BYTE;

//Macros
#define Qstrlen(TheStr) (sizeof(TheStr)-1)					//Quick strlen for constant buffers

//Includes
	//ZLib Includes
	#include "zlib\unzip.h"
	#pragma comment(lib, "./zlib/zlib.lib")

	//Standard stuff
	#include <fcntl.h> //_O_BINARY
	#include <io.h> //_setmode
	#include <stdio.h>
	#include <time.h>

//The info passed to our program to extract
struct RequestedInfo
{
	char *FileName;											//Name of the zip file
	char *FileDownload;										//File to download inside the zip
	char *FullPath;											//Full URL address to zip file
};

inline RequestedInfo GetRequestedInfo(char *env[]);			//Get what info was requested
void OutputFile(int FileSize);								//Writes HTTP headers for a file

int main(int argc, char *argv[], char *env[])
{
	//Initiate variables
	RequestedInfo MyInfo=GetRequestedInfo(env); //Get what info was requested
	int RetVal=S_OK; //Value to return on exit
	_setmode(fileno(stdout), _O_BINARY); //Set stdout to binary
	const char* HTMLHead="Content-Type:"; //Start of HTML string - IIS suggests adding "HTTP/1.0 200 OK\n" to the beggining

	//Try opening the zip file
	unzFile OpenedFile=unzOpen(MyInfo.FileName);
	if(!OpenedFile) //If zip does not exist, output and return error
	{
		printf(
			"%s text/html\n\n"
			"<HTML><BODY>Error, <b>%s</b> does not exist or is invalid.</BODY><HTML>",
			HTMLHead, MyInfo.FileName);
		RetVal=E_FAIL;
		goto EndProcess;
	}

	//Check to see what kind operation was requested, either list files (NULL), download the file (download), or get a file from inside the zip (other)
	if(MyInfo.FileDownload!=NULL)
	{
		if(strcmp(MyInfo.FileDownload, "download")==0) //Download the whole zip
		{
			//Read the data contents
			unzClose(OpenedFile); //Close the zip file first since we just want the binary data
			
			//Get the file size
			FILE *FileNum=fopen(MyInfo.FileName, "rb");
			fseek(FileNum, 0, SEEK_END);
			DWORD FileSize=ftell(FileNum);
			fseek(FileNum, 0, SEEK_SET);

			//Output the headers for the zip file
			printf("%s application/zip\n", HTMLHead);
			OutputFile(FileSize);


			//Read in and print the file 1 megabyte at a time
			const int Megabyte=1024*1024;
			BYTE *Buffer = new BYTE[FileSize<Megabyte?FileSize:Megabyte];
			while(FileSize) //While there is still more to read, read next chunk
			{
				DWORD SizeToRead=(FileSize<Megabyte?FileSize:Megabyte); //Get size of chunk to read
				fread(Buffer, SizeToRead, 1, FileNum); //Read the chunk
				fwrite(Buffer, SizeToRead, 1, stdout); //Write out the chunk
				FileSize-=SizeToRead; //Subtract the chunk size
			}

			//Delete buffers, close file, and return success
			fclose(FileNum);
			delete[] Buffer;
			goto EndProcess;
		}

		//Get view type (save or view)
		int FileDownloadType=*(MyInfo.FileDownload)-'0'; //1st character is either 0 or 1 for view or save
		char *FileDownload=MyInfo.FileDownload+1; //Actual file download location starts after type of download character

		//Check and make sure the file is in the zip and 
		if(unzLocateFile(OpenedFile, FileDownload, false)!=UNZ_OK) //Check to make sure file exists in zip, and if not, output and return error
		{
			printf(
				"%s text/html\n\n"
				"<HTML><BODY>Error, <b>%s</b> is not within <b>%s</b>.</BODY><HTML>",
				HTMLHead, FileDownload, MyInfo.FileName);
			unzClose(OpenedFile);
			RetVal=E_FAIL;
			goto EndProcess;
		}
		
		//Prepare to read file info
		unz_file_info fileinfoIn;
		unzGetCurrentFileInfo(OpenedFile, &fileinfoIn, NULL, 0, NULL, 0, NULL, 0);
		BYTE *Buffer = new BYTE[fileinfoIn.uncompressed_size];
		if(!Buffer) //If we cannot create a buffer large enough to read it, output and return error
		{
			unzCloseCurrentFile(OpenedFile);
		    unzClose(OpenedFile);
			printf(
				"%s text/html\n\n"
				"<HTML><BODY>Error, not able to allocate enough memory for <b>%s</b> within <b>%s</b>.</BODY><HTML>",
				HTMLHead, MyInfo.FileDownload, MyInfo.FileName);
			unzClose(OpenedFile);
			RetVal=E_OUTOFMEMORY;
			goto EndProcess;
		}

		//Write headers
		if(FileDownloadType) //Save file
		{
			char *FileWithoutPath=strrchr(FileDownload, '/'); //Get the filename without the path
			printf("%s application/octet-stream\nContent-disposition: attachment;filename=%s\r\n", HTMLHead, FileWithoutPath ? FileWithoutPath+1 : FileDownload);
		}
		else //View file;
			printf("%s text/html\n", HTMLHead);
		OutputFile(fileinfoIn.uncompressed_size);

		//Read the whole file in, close it, and then write it out
		unzOpenCurrentFile(OpenedFile);
		unzReadCurrentFile(OpenedFile, Buffer, fileinfoIn.uncompressed_size);
		unzCloseCurrentFile(OpenedFile);
		unzClose(OpenedFile);
		fwrite(Buffer, fileinfoIn.uncompressed_size, 1, stdout);

		//Delete buffers return success
		delete[] Buffer;
		goto EndProcess;
	}

	//Prepare to return a table of files and attributes in the zip with download links, with the table header as the zip file link and data itself
	printf(
		"%s text/html\n\n"
		"<HTML><HEAD><TITLE>%s</TITLE><STYLE type=\"text/css\">.MainTable {width:100\\%;}.MainTable td {white-space:nowrap;}</STYLE></HEAD><BODY>Download Full ZIP: <a href=\"http://%s?download\">%s</a><br><br>"
		"<table class=\"MainTable\"><tr><th>Name</th><th>Save</th><th>Date</th><th>Compressed Size</th><th>Uncompressed Size</th></tr>",
		HTMLHead, strrchr(MyInfo.FileName, '\\')+1, MyInfo.FullPath, MyInfo.FullPath);

	//Write out rows in table for files
	unzGoToFirstFile(OpenedFile);
	do
	{
		unz_file_info fileinfo;
		char InFileName[255];
		unzGetCurrentFileInfo(OpenedFile, &fileinfo, InFileName, 255, NULL, 0, NULL, 0);
		printf("<tr><td><a href=\"http://%s?0%s\">%s</a></td><td><a href=\"http://%s?1%s\">Save</a></td><td>%02i/%02i/%04i %02i:%02i:%02i</td><td>%i</td><td>%i</td></tr>",MyInfo.FullPath,InFileName,InFileName,MyInfo.FullPath,InFileName,fileinfo.tmu_date.tm_mon, fileinfo.tmu_date.tm_mday,fileinfo.tmu_date.tm_year,fileinfo.tmu_date.tm_hour, fileinfo.tmu_date.tm_min, fileinfo.tmu_date.tm_sec, fileinfo.compressed_size, fileinfo.uncompressed_size);
	} while(unzGoToNextFile(OpenedFile)!=UNZ_END_OF_LIST_OF_FILE);	
	
	//End table, and close out file and buffers
	printf("</TABLE></BODY></HTML>");
	unzClose(OpenedFile);

	//Clear out variables
	EndProcess:
	delete[] MyInfo.FileName;
	delete[] MyInfo.FileDownload;
	delete[] MyInfo.FullPath;

	return S_OK;
}

inline RequestedInfo GetRequestedInfo(char *env[]) //Get what info was requested
{
	//Strings to compare for common HTTP headers, this way strlen's are auto calculated
	const char PathToFile[] = "PATH_TRANSLATED=";
	const char Parameters[] = "QUERY_STRING="; //
	const char Host[] = "HTTP_HOST="; //Host and info are concatenated to create FullPath
	const char Path[] = "PATH_INFO=";

	RequestedInfo MyInfo={NULL,NULL,NULL}; //Returned/collected info
	for(int count=0;env[count];count++) //Loop until zero delimiter is hit
	{
		if(strncmp(PathToFile, env[count], Qstrlen(PathToFile))==0) //File path
		{
			delete[] MyInfo.FileName; //If already defined, delete old pointer
			MyInfo.FileName=new char[strlen(env[count])-Qstrlen(PathToFile)+1];
			strcpy(MyInfo.FileName, env[count]+Qstrlen(PathToFile));
		}

		if(strncmp(Parameters, env[count], Qstrlen(Parameters))==0) //Parameter string
		{
			if(Qstrlen(Parameters)<strlen(env[count])) //If there are any parameters...
			{
				delete[] MyInfo.FileDownload; //If already defined, delete old pointer
				MyInfo.FileDownload=new char[strlen(env[count])-Qstrlen(Parameters)+1];
				strcpy(MyInfo.FileDownload, env[count]+Qstrlen(Parameters));
				while(strstr(MyInfo.FileDownload, "%20")) //Replace " " with "%20"
				{
					char* TmpStr=strstr(MyInfo.FileDownload, "%20");
					strcpy(TmpStr+1, TmpStr+3);
					*TmpStr=' ';
				}
			}
		}

		for(int i=0;i<2;i++) //Process the Path and Host variables, making sure the FullPath is Host+Path
			if(strncmp(i ? Path : Host, env[count], (i?Qstrlen(Path):Qstrlen(Host)))==0) //Check if we are on Path or Host
			{
				char *MyStr=env[count]+(i?Qstrlen(Path):Qstrlen(Host)); //Get the pointer to the actual variable information
				DWORD NewLen=strlen(MyStr)+(!MyInfo.FullPath ? 0 : strlen(MyInfo.FullPath)); //Length of the output
				char *FinalPath=new char[NewLen+1]; //Create the output character string
				if(!MyInfo.FullPath) //If we do not already have info, just copy in the new info 
					strcpy(FinalPath, MyStr);
				else //Otherwise we have to concatenate the strings in the right order
				{
					strcpy(FinalPath, i ? MyInfo.FullPath : MyStr);
					strcat(FinalPath, i ? MyStr : MyInfo.FullPath);
					delete[] MyInfo.FullPath;
				}
				MyInfo.FullPath=FinalPath; //Remember our concatenated string
			}
	}
	return MyInfo;
}

void OutputFile(int FileSize) //Writes HTTP headers for a file
{
	//Create a time string
	char DateAndTime[255];
	time_t ltime;
	time(&ltime);
	struct tm *today = localtime(&ltime);
	strftime(DateAndTime, 255, "%a, %d %b %Y %H:%M:%S %Z", today);

	//Write out HTTP headers for the file
	printf(
		"Date: %s\r\n"
		"Server: Microsoft-IIS/5.1\n"
		"Accept-Ranges: bytes\n"
		"Content-Length: %d\r\n"
		"Connection: close\n\n",
		DateAndTime, FileSize);
}